from typing import List, TypeVar, Generic

T = TypeVar("T")


class PseudoFrozenList(Generic[T]):
    # we don't need  python list's all methods, so don't inherit from it
    @staticmethod  # noqa
    def __class_getitem__(*types_: T):
        """
        Another initialization method. I just liked it this way :>
        >>> x = PseudoFrozenList[int]
        >>> x.append(1, 2, 3, "4,5,6,7".split(","))
        :return: new PseudoFrozenList
        """
        return PseudoFrozenList(*types_)

    def __init__(self, t: T):
        if not isinstance(t, type):
            raise ValueError(f"Expected Type, got {t!r}")

        self.__items: List[T] = list()
        self.__frozen = False

    @property
    def frozen(self):
        return self.__frozen

    def append(self, *items: T) -> None:
        ap = self.__items.append
        for val in items:
            ap(val)

    def remove(self, *items: T) -> None:
        rm = self.__items.remove
        for val in items:
            rm(val)

    def freeze(self):
        self.__frozen = True

    def __iter__(self):
        return list.__iter__(self.__items)

    # decorator for append
    @property
    def __call__(self):
        def wrp(item: T):
            self.append(item)

        return wrp



from __future__ import annotations
import operator
import inspect
from typing import Callable, Any


class Filter:
    """
    Base Filter object, callback container
    """

    def __init__(
        self, function: Callable, requires_context: bool = False, state_op: bool = False
    ):
        if isinstance(function, Filter):
            self.__name__ = function.__qualname__
            self.function = function.function
            self.awaitable = function.awaitable
            self.requires_context = function.requires_context
            self.key_maker = function.key_maker

        else:
            name = function.__qualname__
            if name == "<lambda>":
                name = f"{function!s}"

            self.__name__ = f"Garnet:Filter:{name}"

            self.function = function
            self.awaitable = inspect.iscoroutinefunction(
                function
            ) or inspect.isawaitable(function)
            self.requires_context = requires_context

        self.key_maker = None

        self.__qualname__ = self.__name__
        self.state_op = state_op

    @classmethod
    def from_def_state(cls, state_op: bool):
        return lambda function, requires_context=False: cls(
            function, requires_context, state_op
        )

    def __eq__(self, value: Any) -> Filter:
        return _singfilter(self, lambda filter1: operator.eq(filter1, value))

    def __ne__(self, value: Any) -> Filter:
        return _singfilter(self, lambda filter1: operator.ne(filter1, value))

    def __xor__(self, filter2: Filter) -> Filter:
        return _compose_filter(self, filter2, operator.xor)

    def __and__(self, filter2: Filter) -> Filter:
        return _compose_filter(self, filter2, operator.and_)

    def __or__(self, filter2: Filter) -> Filter:
        return _compose_filter(self, filter2, operator.or_)

    def __invert__(self) -> Filter:
        return _singfilter(self, operator.not_)


def _compose_filter(filter1: Filter, filter2: Filter, operator_) -> Filter:
    if (not isinstance(filter1, Filter)) | (not isinstance(filter2, Filter)):
        raise ValueError(
            f"Cannot compare non-Filter object with Filter, "
            f"got filter1={filter1.__qualname__} "
            f"operator={operator_.__qualname__} "
            f"filter2={filter2.__qualname__}"
        )

    result = Filter.from_def_state(filter1.state_op | filter2.state_op)
    if (filter1.awaitable & filter2.awaitable) & (
        filter1.requires_context & filter2.requires_context
    ):

        async def func(event, context):
            return operator_(
                await filter1.function(event, context),
                await filter2.function(event, context),
            )

        return result(func, True)

    elif (filter1.awaitable ^ filter2.awaitable) & (
        filter1.requires_context ^ filter2.requires_context
    ):
        # as told, passing context to sync function has no sense
        # so awaitable function here is which requires context
        is_f1_async = filter1.awaitable or filter1.requires_context
        async_func = filter1 if is_f1_async else filter2
        sync_func = filter2 if is_f1_async else filter1

        async def func(event, context):
            return operator_(
                await async_func.function(event, context), sync_func.function(event)
            )

        return result(func, True)

    elif filter1.awaitable ^ filter2.awaitable:
        async_func = filter1 if filter1.awaitable else filter2
        sync_func = filter2 if async_func == filter1 else filter1

        async def func(event):
            return operator_(
                await async_func.function(event), sync_func.function(event)
            )

        return result(func)

    def func(event):
        return operator_(filter1.function(event), filter2.function(event))

    return result(func)


def _singfilter(filter1: Filter, operator_) -> Filter:
    result = Filter.from_def_state(filter1.state_op)

    if filter1.awaitable | filter1.requires_context:
        if filter1.requires_context:

            async def func(event, context):
                return operator_(await filter1.function(event, context))

            return result(func, requires_context=True)

        async def func(event):
            return operator_(await filter1.function(event))

        return result(func)

    def func(event):
        return operator_(filter1.function(event))

    return result(func)


__all__ = ("Filter",)



import inspect
from typing import Callable, Union

from ..filters.base import Filter

CALLBACK_STATE_KEY = "context"
CURRENT_CLIENT_KEY = "current_client"


class Callback:
    __call__ = None

    def __init__(
        self,
        callback: Union[Callable, "Callback"] = None,
        *filters: Union[Callable, Filter]
    ):
        self.filters = filters
        self.requires_context = False
        self.requires_client = False
        self.key_maker = None
        self.continue_prop = False
        self.empty_arg_handler = False

        if isinstance(callback, Callable):
            self.__call__ = callback

        elif isinstance(callback, Callback):
            callback = callback.__call__

        if not isinstance(self.__call__, Callable):
            # .__call__ must be set by inherited class, otherwise Error
            raise ValueError("Callback must be callable")

        args = inspect.getfullargspec(callback)
        args_cp = args.args.copy()
        args_cp.extend(args.kwonlyargs)

        if not args_cp:
            self.empty_arg_handler = True

        else:
            if CALLBACK_STATE_KEY in args_cp:
                self.requires_context = True

            if CURRENT_CLIENT_KEY in args_cp:
                self.requires_client = True



from telethon.utils import is_audio, is_gif, is_image, is_video

from ..helpers import var
from .base import Filter


class _FilterVar(var.Var):
    def __set_name__(self, owner, val):
        if not self.value:
            self.value = Filter(
                lambda event: event.file is not None
                and (self.prefix + val + self.suffix) == event.file.ext
            )


class File:
    is_video: Filter = Filter(lambda event: is_video(event.media))
    is_image: Filter = Filter(lambda event: is_image(event.media))
    is_gif: Filter = Filter(lambda event: is_gif(event.document))
    is_audio: Filter = Filter(lambda event: is_audio(event.media))

    # generated by extensions
    # Image extensions
    png: Filter = _FilterVar()
    jpeg: Filter = _FilterVar()
    webp: Filter = _FilterVar()
    gif: Filter = _FilterVar()
    bmp: Filter = _FilterVar()
    tga: Filter = _FilterVar()
    tiff: Filter = _FilterVar()
    psd: Filter = _FilterVar()

    # Video extensions
    mp4: Filter = _FilterVar()
    mov: Filter = _FilterVar()
    avi: Filter = _FilterVar()

    # Audio extensions
    mp2: Filter = _FilterVar()  # a.k.a. ISO/IEC 13818-2
    mp3: Filter = _FilterVar()
    m4a: Filter = _FilterVar()
    aac: Filter = _FilterVar()
    ogg: Filter = _FilterVar()
    flac: Filter = _FilterVar()

    # Win executable extension
    exe: Filter = _FilterVar()



from typing import List, Callable, Dict, Any, Tuple, Union, Sequence

from telethon.events import common

from .base import Filter
from ..storages.base import FSMContext

ANY_STATE = (id(all), id(any))
KeyMakerType = Callable[[common.EventBuilder], Dict[str, Any]]


class _MetaCurrentState(type):
    @classmethod
    def __matmul__(mcs, key_maker: Tuple[str, KeyMakerType]) -> Filter:
        equal_to_state, key_maker = key_maker

        async def _f(event, context: FSMContext):
            # any context can be passed, just a small hack
            return await context.storage.get_state(**key_maker(event)) == equal_to_state

        f_ = Filter(_f, requires_context=True)
        f_.key_maker = key_maker
        return f_

    @classmethod
    def __rmatmul__(mcs, key_maker: Tuple[str, KeyMakerType]) -> Filter:
        return CurrentState @ key_maker

    @classmethod
    def __eq__(mcs, _s: Union[str, List[str]]) -> Filter:
        if id(_s) in ANY_STATE:

            async def _f(_, context):
                return (await context.get_state()) is not None

        elif isinstance(_s, Sequence):

            async def _f(_, context):
                return (await context.get_state()) in _s

        else:

            async def _f(_, context):
                return (await context.get_state()) == _s

        return Filter(_f, requires_context=True, state_op=True)

    @classmethod
    def exact(mcs, state: Union[str, List[str]]) -> Filter:
        # noinspection PyTypeChecker
        return CurrentState == state  # type: ignore

    @classmethod
    def with_key(mcs, state: str, key_maker: KeyMakerType):
        return CurrentState @ (state, key_maker)


class CurrentState(metaclass=_MetaCurrentState):
    """
    Class with implemented in metaclass magic methods

    CurrentState @ ("state_name", key_maker)
        key_maker is a callable with signature: (event) -> Dict[str, Any]
        For instance:
            >>> def my_key_maker(event) -> Dict[str, Any]:
            ...     return {"chat": event.chat.id, event.user.id}
            ...
            >>> CurrentState @ ("equlas_to_the_state", my_key_maker)
            ... <garnet.filters.base.Filter object at 0x7f8bebe996>
            >>> # same approaching from the right side.
            >>> ("equlas_to_the_state", my_key_maker) @ CurrentState
            ... <garnet.filters.base.Filter object at 0x7f8baba001>

    Any state: ANY_STATE states "*". Simply use equality operator to builtin methods
        CurrentState == all
        CurrentState == any

    Particular state(s):
        CurrentState == ["state_1", "state_2"]
        CurrentState == "state_1"

    """


__all__ = ("CurrentState",)



import logging
import json as _std_json
import importlib

json = _std_json

for __json_lib in ("orjson", "ujson", "rapidjson"):
    try:
        json = importlib.import_module(__json_lib)
        dumps = json.dumps
        loads = json.loads
        logging.info(f"{__name__}: Found {__json_lib} installed. Going to use it.")
        break
    except ImportError:
        continue



from __future__ import annotations
import typing

# noinspection PyPackageRequirements
import contextvars

Type = typing.TypeVar("Type")


class ContextInstanceMixin:
    def __init_subclass__(cls, **kwargs):
        cls.__ctx_var = contextvars.ContextVar("initialized_" + cls.__name__)
        return cls

    @classmethod
    def current(cls: typing.Type[Type], no_error=True) -> Type:
        if no_error:
            return cls.__ctx_var.get(None)
        return cls.__ctx_var.get()

    @classmethod
    def set_current(cls: typing.Type[Type], value: Type) -> typing.NoReturn:
        if not isinstance(value, cls):
            raise ValueError(f"Got wrong instance, expected {cls!r} got {value!r}")

        cls.__ctx_var.set(value)



import operator
import typing
import re

from .base import Filter


def _base_len_comparator(operator_, predicted_length):
    if predicted_length is None:
        return Filter(lambda update: operator_(len(update.raw_text or "")))
    else:
        return Filter(
            lambda update: operator_(len(update.raw_text or ""), predicted_length)
        )


def startswith(prefix: str) -> Filter:
    """
    Check if .event.raw_text startswith prefix
    :return:
    """
    return Filter(
        lambda update: update.raw_text.startswith(prefix)
        if isinstance(update.raw_text, str)
        else None
    )


def commands(
    *cmd: str, prefixes: typing.Union[str, typing.Iterable[str]] = "/"
) -> Filter:
    return Filter(
        lambda update: isinstance(update.text, str)
        and any(update.text.startswith(prefix) for prefix in prefixes)
        and update.text.split()[0][1:] in cmd
    )


def match(expression: str) -> Filter:
    return Filter(lambda update: re.compile(expression).match(update.raw_text))


def exact(text: str) -> Filter:
    return Filter(lambda update: update.raw_text == text)


def between(*texts: str) -> Filter:
    return Filter(lambda update: update.raw_text in texts)


def isdigit() -> Filter:
    return Filter(lambda update: (update.raw_text or "").isdigit())


class MessageLenMeta(type):
    def __eq__(self, length: int) -> Filter:
        return _base_len_comparator(operator.eq, length)

    def __gt__(self, length: int) -> Filter:
        return _base_len_comparator(operator.gt, length)

    def __lt__(self, length: int) -> Filter:
        return _base_len_comparator(operator.lt, length)

    def __ge__(self, length: int) -> Filter:
        return _base_len_comparator(operator.ge, length)

    def __le__(self, length: int) -> Filter:
        return _base_len_comparator(operator.le, length)


class Len(metaclass=MessageLenMeta):
    """
    Text Length
    """



class Var:
    """
    Set string object values of class attributes as their link names.
    Pulled from https://github.com/uwinx/tamtam.py [modified]
    """

    def __init__(self, prefix: str = "", value=None, suffix: str = ""):
        self.value = value
        self.prefix = prefix
        self.suffix = suffix

    def __set_name__(self, owner, val):
        if not self.value:
            self.value = self.prefix + val + self.suffix

    def __get__(self, instance, owner):
        return self.value

    def __str__(self):
        return self.value.__str__()



"""
garnet.Boost was ported from t*lir's one lib
"""
import asyncio
import logging
import math
from typing import Optional, List, AsyncGenerator, Union, Awaitable, Tuple

from telethon import utils
from telethon.crypto import AuthKey
from telethon.network import MTProtoSender
from telethon.tl.functions.auth import (
    ExportAuthorizationRequest,
    ImportAuthorizationRequest,
)
from telethon.tl.functions.upload import (
    GetFileRequest,
    SaveFilePartRequest,
    SaveBigFilePartRequest,
)
from telethon.tl.types import (
    Document,
    InputFileLocation,
    InputDocumentFileLocation,
    InputPhotoFileLocation,
    InputPeerPhotoFileLocation,
)
from . import TelegramClient

log = logging.getLogger("garnet.boost")

TypeLocation = Union[
    Document,
    InputDocumentFileLocation,
    InputPeerPhotoFileLocation,
    InputFileLocation,
    InputPhotoFileLocation,
]

MAX_FILE_SIZE = 100 * 1024 * 1024


class DownloadSender:
    def __init__(
        self,
        sender: MTProtoSender,
        file: TypeLocation,
        offset: int,
        limit: int,
        stride: int,
        count: int,
    ) -> None:
        log.debug(f"Creating download sender with {offset} {limit} {stride} {count}")
        self.sender = sender
        self.request = GetFileRequest(file, offset=offset, limit=limit)
        self.stride = stride
        self.remaining = count

    async def next(self) -> Optional[bytes]:
        if not self.remaining:
            return None
        result = await self.sender.send(self.request)
        self.remaining -= 1
        self.request.offset += self.stride
        return result.bytes

    def disconnect(self) -> Awaitable[None]:
        return self.sender.disconnect()


class UploadSender:
    def __init__(
        self,
        sender: MTProtoSender,
        file_id: int,
        part_count: int,
        big: bool,
        index: int,
        stride: int,
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        log.debug(
            f"Creating upload sender with {file_id} {part_count} {big} {index} {stride}"
        )
        self.sender = sender
        self.part_count = part_count
        if big:
            self.request = SaveBigFilePartRequest(file_id, index, part_count, b"")
        else:
            self.request = SaveFilePartRequest(file_id, index, b"")
        self.stride = stride
        self.previous = None
        self.loop = loop

    async def next(self, data: bytes) -> None:
        if self.previous:
            await self.previous
        self.previous = self.loop.create_task(self._next(data))

    async def _next(self, data: bytes) -> None:
        self.request.bytes = data
        log.debug(
            f"Sending file part {self.request.file_part}/{self.part_count} "
            f"with {len(data)} bytes"
        )
        await self.sender.send(self.request)
        self.request.file_part += self.stride

    async def disconnect(self) -> None:
        if self.previous:
            await self.previous
        return await self.sender.disconnect()


class ParallelTransferrer:
    loop: asyncio.AbstractEventLoop
    dc_id: int
    senders: Optional[List[Union[DownloadSender, UploadSender]]]
    auth_key: AuthKey
    upload_ticker: int

    def __init__(self, client: TelegramClient, dc_id: Optional[int] = None) -> None:
        self.client = client
        self.loop = self.client.loop
        self.dc_id = dc_id or self.client.session.dc_id
        self.auth_key = (
            None
            if dc_id and self.client.session.dc_id != dc_id
            else self.client.session.auth_key
        )
        self.senders = None
        self.upload_ticker = 0

    async def _cleanup(self) -> None:
        await asyncio.gather(*[sender.disconnect() for sender in self.senders])
        self.senders = None

    @staticmethod
    def _get_connection_count(
        file_size: int, max_count: int = 20, full_size: int = MAX_FILE_SIZE
    ) -> int:
        if file_size > full_size:
            return max_count
        return math.ceil((file_size / full_size) * max_count)

    async def _init_download(
        self, connections: int, file: TypeLocation, part_count: int, part_size: int
    ) -> None:
        minimum, remainder = divmod(part_count, connections)

        def get_part_count() -> int:
            nonlocal remainder
            if remainder > 0:
                remainder -= 1
                return minimum + 1
            return minimum

        # The first cross-DC sender will export+import the authorization, so we always create it
        # before creating any other senders.
        self.senders = [
            await self._create_download_sender(
                file, 0, part_size, connections * part_size, get_part_count()
            ),
            *await asyncio.gather(
                *[
                    self._create_download_sender(
                        file, i, part_size, connections * part_size, get_part_count()
                    )
                    for i in range(1, connections)
                ]
            ),
        ]

    async def _create_download_sender(
        self,
        file: TypeLocation,
        index: int,
        part_size: int,
        stride: int,
        part_count: int,
    ) -> DownloadSender:
        return DownloadSender(
            await self._create_sender(),
            file,
            index * part_size,
            part_size,
            stride,
            part_count,
        )

    async def _init_upload(
        self, connections: int, file_id: int, part_count: int, big: bool
    ) -> None:
        self.senders = [
            await self._create_upload_sender(file_id, part_count, big, 0, connections),
            *await asyncio.gather(
                *[
                    self._create_upload_sender(file_id, part_count, big, i, connections)
                    for i in range(1, connections)
                ]
            ),
        ]

    async def _create_upload_sender(
        self, file_id: int, part_count: int, big: bool, index: int, stride: int
    ) -> UploadSender:
        return UploadSender(
            await self._create_sender(),
            file_id,
            part_count,
            big,
            index,
            stride,
            loop=self.loop,
        )

    async def _create_sender(self) -> MTProtoSender:
        dc = await self.client._get_dc(self.dc_id)
        sender = MTProtoSender(self.auth_key, self.loop, loggers=self.client._log)
        await sender.connect(
            self.client._connection(
                dc.ip_address,
                dc.port,
                dc.id,
                loop=self.loop,
                loggers=self.client._log,
                proxy=self.client._proxy,
            )
        )
        if not self.auth_key:
            log.debug(f"Exporting auth to DC {self.dc_id}")
            auth = await self.client(ExportAuthorizationRequest(self.dc_id))
            req = self.client._init_with(
                ImportAuthorizationRequest(id=auth.id, bytes=auth.bytes)
            )
            await sender.send(req)
            self.auth_key = sender.auth_key
        return sender

    async def init_upload(
        self,
        file_id: int,
        file_size: int,
        part_size_kb: Optional[float] = None,
        connection_count: Optional[int] = None,
    ) -> Tuple[int, int, bool]:
        connection_count = connection_count or self._get_connection_count(file_size)
        part_size = (part_size_kb or utils.get_appropriated_part_size(file_size)) * 1024
        part_count = (file_size + part_size - 1) // part_size
        is_large = file_size > MAX_FILE_SIZE // 10
        await self._init_upload(connection_count, file_id, part_count, is_large)
        return part_size, part_count, is_large

    async def upload(self, part: bytes) -> None:
        await self.senders[self.upload_ticker].next(part)
        self.upload_ticker = (self.upload_ticker + 1) % len(self.senders)

    async def finish_upload(self) -> None:
        await self._cleanup()

    async def download(
        self,
        file: TypeLocation,
        file_size: int,
        part_size_kb: Optional[float] = None,
        connection_count: Optional[int] = None,
    ) -> AsyncGenerator[bytes, None]:
        connection_count = connection_count or self._get_connection_count(file_size)
        part_size = (part_size_kb or utils.get_appropriated_part_size(file_size)) * 1024
        part_count = math.ceil(file_size / part_size)
        log.debug(
            f"Starting parallel download: "
            f"{connection_count} {part_size} {part_count} {file!s}"
        )
        await self._init_download(connection_count, file, part_count, part_size)

        part = 0
        while part < part_count:
            tasks = []
            for sender in self.senders:
                tasks.append(self.loop.create_task(sender.next()))
            for task in tasks:
                data = await task
                if not data:
                    break
                yield data
                part += 1
                log.debug(f"Part {part} downloaded")

        log.debug("Parallel download finished, cleaning up connections")
        await self._cleanup()



import typing
import asyncio
from abc import ABC

import aiomcache

from tgarnet.storages.base import BaseStorage


class MemcachedStorage(BaseStorage, ABC):
    def __init__(self, host: str, port: int, loop: asyncio.BaseEventLoop = None):
        self.__cache = aiomcache.Client(**locals())

    async def close(self):
        await self.__cache.close()

    async def get_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Optional[str]:
        await self.__cache.get(b"")


async def hello_aiomcache():
    mc = aiomcache.Client("127.0.0.1", 11211)
    await mc.set(b"some_key", b"Some value")
    value = await mc.get(b"some_key")
    print(value)
    values = await mc.multi_get(b"some_key", b"other_key")
    print(values)
    await mc.delete(b"another_key")



# Source: https://github.com/aiogram/aiogram

import copy
import typing
from warnings import warn

# Leak bucket
KEY = "key"


class BaseStorage:
    """
    You are able to save current user's state
    and data for all steps in states-storage
    """

    async def close(self):
        """
        You have to override this method and use when application shutdowns.
        Perhaps you would like to save data and etc.

        :return:
        """
        raise NotImplementedError

    async def wait_closed(self):
        """
        You have to override this method for all asynchronous storages (e.g., Redis).

        :return:
        """
        raise NotImplementedError

    @classmethod
    def check_address(
        cls,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
    ) -> (typing.Union[str, int], typing.Union[str, int]):
        """
        In all storage's methods chat or user is always required.
        If one of them is not provided, you have to set missing value based on the provided one.

        This method performs the check described above.

        :param chat:
        :param user:
        :return:
        """
        if chat is None and user is None:
            raise ValueError(
                "`user` or `chat` parameter is required but no one is provided!"
            )

        if user is None and chat is not None:
            user = chat
        elif user is not None and chat is None:
            chat = user
        return chat, user

    async def get_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Optional[str]:
        """
        Get current state of user in chat. Return `default` if no record is found.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :param default:
        :return:
        """
        raise NotImplementedError

    async def get_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[typing.Dict] = None,
    ) -> typing.Dict:
        """
        Get state-data for user in chat. Return `default` if no data is provided in storage.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :param default:
        :return:
        """
        raise NotImplementedError

    async def set_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        state: typing.Optional[typing.AnyStr] = None,
    ):
        """
        Set new state for user in chat

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :param state:
        """
        raise NotImplementedError

    async def set_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
    ):
        """
        Set data for user in chat

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :param data:
        """
        raise NotImplementedError

    async def update_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
        **kwargs,
    ):
        """
        Update data for user in chat

        You can use data parameter or|and kwargs.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param data:
        :param chat:
        :param user:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    async def reset_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
    ):
        """
        Reset data for user in chat.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :return:
        """
        await self.set_data(chat=chat, user=user, data={})

    async def reset_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        with_data: typing.Optional[bool] = True,
    ):
        """
        Reset state for user in chat.
        You may desire to use this method when finishing conversations.

        Chat or user is always required. If one of this is not presented,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :param with_data:
        :return:
        """
        chat, user = self.check_address(chat=chat, user=user)
        await self.set_state(chat=chat, user=user, state=None)
        if with_data:
            await self.set_data(chat=chat, user=user, data={})

    async def finish(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
    ):
        """
        Finish conversation for user in chat.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :return:
        """
        await self.reset_state(chat=chat, user=user, with_data=True)

    def has_bucket(self):
        return False

    async def get_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[dict] = None,
    ) -> typing.Dict:
        """
        Get bucket for user in chat. Return `default` if no data is provided in storage.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :param default:
        :return:
        """
        raise NotImplementedError

    async def set_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None,
    ):
        """
        Set bucket for user in chat

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :param bucket:
        """
        raise NotImplementedError

    async def update_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None,
        **kwargs,
    ):
        """
        Update bucket for user in chat

        You can use bucket parameter or|and kwargs.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param bucket:
        :param chat:
        :param user:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    async def reset_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
    ):
        """
        Reset bucket dor user in chat.

        Chat or user is always required. If one of them is not provided,
        you have to set missing value based on the provided one.

        :param chat:
        :param user:
        :return:
        """
        await self.set_data(chat=chat, user=user, data={})


class FSMContext:
    def __init__(self, storage, chat, user):
        self.storage: BaseStorage = storage
        self.chat, self.user = self.storage.check_address(chat=chat, user=user)

    def proxy(self):
        return FSMContextProxy(self)

    async def get_state(self) -> typing.Optional[str]:
        return await self.storage.get_state(chat=self.chat, user=self.user)

    async def get_data(self) -> typing.Dict:
        return await self.storage.get_data(chat=self.chat, user=self.user)

    async def update_data(self, data: typing.Dict = None, **kwargs):
        await self.storage.update_data(
            chat=self.chat, user=self.user, data=data, **kwargs
        )

    async def set_state(self, state: typing.Union[typing.AnyStr, None] = None):
        await self.storage.set_state(chat=self.chat, user=self.user, state=state)

    async def set_data(self, data: typing.Dict = None):
        await self.storage.set_data(chat=self.chat, user=self.user, data=data)

    async def reset_state(self, with_data: typing.Optional[bool] = True):
        await self.storage.reset_state(
            chat=self.chat, user=self.user, with_data=with_data
        )

    async def reset_data(self):
        await self.storage.reset_data(chat=self.chat, user=self.user)

    async def finish(self):
        await self.storage.finish(chat=self.chat, user=self.user)


class FSMContextProxy:
    def __init__(self, fsm_context: FSMContext):
        super(FSMContextProxy, self).__init__()
        self.fsm_context = fsm_context
        self._copy = {}
        self._data = {}
        self._state = None
        self._is_dirty = False

        self._closed = True

    async def __aenter__(self):
        await self.load()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            await self.save()
        self._closed = True

    def _check_closed(self):
        if self._closed:
            raise LookupError("Proxy is closed!")

    @classmethod
    async def create(cls, fsm_context: FSMContext):
        """
        :param fsm_context:
        :return:
        """
        proxy = cls(fsm_context)
        await proxy.load()
        return proxy

    async def load(self):
        self._closed = False

        self.clear()
        self._state = await self.fsm_context.get_state()
        self.update(await self.fsm_context.get_data())
        self._copy = copy.deepcopy(self._data)
        self._is_dirty = False

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, value):
        self._check_closed()

        self._state = value
        self._is_dirty = True

    @state.deleter
    def state(self):
        self._check_closed()

        self._state = None
        self._is_dirty = True

    async def save(self, force=False):
        self._check_closed()

        if self._copy != self._data or force:
            await self.fsm_context.set_data(data=self._data)
        if self._is_dirty or force:
            await self.fsm_context.set_state(self.state)
        self._is_dirty = False
        self._copy = copy.deepcopy(self._data)

    def clear(self):
        del self.state
        return self._data.clear()

    def get(self, value, default=None):
        return self._data.get(value, default)

    def setdefault(self, key, default):
        self._check_closed()

        self._data.setdefault(key, default)

    def update(self, data=None, **kwargs):
        self._check_closed()

        self._data.update(data, **kwargs)

    def pop(self, key, default=None):
        self._check_closed()

        return self._data.pop(key, default)

    def keys(self):
        return self._data.keys()

    def values(self):
        return self._data.values()

    def items(self):
        return self._data.items()

    def as_dict(self):
        return copy.deepcopy(self._data)

    def __len__(self):
        return len(self._data)

    def __iter__(self):
        return self._data.__iter__()

    def __getitem__(self, item):
        return self._data[item]

    def __setitem__(self, key, value):
        self._check_closed()

        self._data[key] = value

    def __delitem__(self, key):
        self._check_closed()

        del self._data[key]

    def __contains__(self, item):
        return item in self._data

    def __str__(self):
        readable_state = f"'{self.state}'" if self.state else "<default>"
        result = (
            f"{self.__class__.__name__} state = {readable_state}, data = {self._data}"
        )
        if self._closed:
            result += ", closed = True"
        return result


class DisabledStorage(BaseStorage):
    """
    Empty storage. Use it if you don't want to use Finite-State Machine
    """

    async def close(self):
        pass

    async def wait_closed(self):
        pass

    async def get_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Optional[str]:
        return None

    async def get_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Dict:
        self._warn()
        return {}

    async def update_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
        **kwargs,
    ):
        self._warn()

    async def set_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        state: typing.Optional[typing.AnyStr] = None,
    ):
        self._warn()

    async def set_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
    ):
        self._warn()

    @staticmethod
    def _warn():
        warn(
            message=(
                f"You haven’t set any storage yet so no states and no data will be saved. \n"
                "You can connect MemoryStorage for debug purposes or non-essential data."
            ),
            category=Warning,
            stacklevel=5,
        )



from typing import Type

from warnings import warn

from telethon.events.raw import Raw
from telethon.events.album import Album
from telethon.events.chataction import ChatAction
from telethon.events.messagedeleted import MessageDeleted
from telethon.events.messageedited import MessageEdited
from telethon.events.messageread import MessageRead
from telethon.events.newmessage import NewMessage
from telethon.events.userupdate import UserUpdate
from telethon.events.callbackquery import CallbackQuery
from telethon.events.inlinequery import InlineQuery

from ..helpers import ctx as _ctx


def patch_event_ctx():
    def ctx_proxy(cls: Type) -> Type:
        def __init__(self, *args, **kwargs):
            cls.__init__(self, *args, **kwargs)
            self.set_current(self)

        return type(
            cls.__name__, (cls, _ctx.ContextInstanceMixin), {"__init__": __init__}
        )

    Album.Event = ctx_proxy(Album.Event)
    ChatAction.Event = ctx_proxy(ChatAction.Event)
    MessageDeleted.Event = ctx_proxy(MessageDeleted.Event)
    MessageEdited.Event = ctx_proxy(MessageEdited.Event)
    MessageRead.Event = ctx_proxy(MessageRead.Event)
    NewMessage.Event = ctx_proxy(NewMessage.Event)
    UserUpdate.Event = ctx_proxy(UserUpdate.Event)
    CallbackQuery.Event = ctx_proxy(CallbackQuery.Event)
    InlineQuery.Event = ctx_proxy(InlineQuery.Event)


patch_event_ctx()


class EventBuilderDict:
    """
    Helper "dictionary" to return events from types and cache them.
    """
    from telethon.events.common import EventCommon  # noqa

    def __init__(self, client, update, others):
        self.client = client
        self.update = update
        self.others = others

    def __getitem__(self, builder):
        try:
            return self.__dict__[builder]
        except KeyError:
            event = self.__dict__[builder] = builder.build(
                self.update, self.others, self.client._self_input_peer.user_id
            )

            if isinstance(event, EventBuilderDict.EventCommon):
                event.original_update = self.update
                event._entities = self.update._entities
                event._set_client(self.client)
            elif event:
                event._client = self.client

            return event


__all__ = (
    "Raw",
    "Album",
    "ChatAction",
    "MessageDeleted",
    "MessageEdited",
    "MessageRead",
    "NewMessage",
    "UserUpdate",
    "CallbackQuery",
    "InlineQuery",
    "EventBuilderDict",
    "StopPropagation",
)


def _deprecated():
    warn(
        message="event.Register/Unregister is deprecated in "
        "Garnet, use garnet::router::Router",
        category=Exception,
        stacklevel=5,
    )


class StopPropagation(Exception):
    """
    If this exception is raised in any of the handlers for a given event,
    it will stop the execution of all other registered event handlers.
    It can be seen as the ``StopIteration`` in a for loop but for events.

    Example usage:

        >>> from telethon import TelegramClient, events
        >>> client = TelegramClient(...)
        >>>
        >>> @client.on(events.NewMessage)
        ... async def delete(event):
        ...     await event.delete()
        ...     # No other event handler will have a chance to handle this event
        ...     raise StopPropagation
        ...
        >>> @client.on(events.NewMessage)
        ... async def _(event):
        ...     # Will never be reached, because it is the second handler
        ...     pass
    """

    # For some reason Sphinx wants the silly >>> or
    # it will show warnings and look bad when generated.
    pass


def register(event=None):
    """
    DEPRECATED FUNCTION
    """
    _deprecated()


def unregister(*_):
    """
    Inverse operation of `register` (though not a decorator). Client-less
    `remove_event_handler
    <telethon.client.updates.UpdateMethods.remove_event_handler>`
    variant. **Note that this won't remove handlers from the client**,
    because it simply can't, so you would generally use this before
    adding the handlers to the client.

    This method is here for symmetry. You will rarely need to
    unregister events, since you can simply just not add them
    to any client.

    If no event is given, all events for this callback are removed.
    Returns how many callbacks were removed.
    """
    _deprecated()


def is_handler(*_):
    """
    DEPRECATED
    Returns `True` if the given callback is an
    event handler (i.e. you used `register` on it).
    """
    _deprecated()


def list(*_):
    """
    DEPRECATED
    Returns a list containing the registered event
    builders inside the specified callback handler.
    """

    _deprecated()


def _get_handlers(*_):
    """
    DEPRECATED
    Like ``list`` but returns `None` if the callback was never registered.
    """
    _deprecated()



from abc import ABC, abstractmethod
from typing import Union, Callable, NoReturn, Sequence, TYPE_CHECKING

from ..events import CallbackQuery, ChatAction, NewMessage, MessageEdited
from ..helpers.frozen_list import PseudoFrozenList
from ..filters import Filter
from ..callbacks.base import Callback

if TYPE_CHECKING:
    from telethon.events import common


class AbstractRouter(ABC):
    frozen: bool
    handlers: Sequence

    @abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def on(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def register(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def non_blocking(self, *args, **kwargs):
        """
        Register handlers in non-blocking way.
        Such as event goes through all `non-blocking` handlers one-by-one ignoring !ALL! exceptions:
        telegram -> update -> execute_all([handler1(update), handler2(update)]) -> happiness and uncertainty

        whereas normal behaviour:
        telegram -> update -> execute_till_filter_success([handler1(update), handler2(update)]) -> just happiness
        """
        raise NotImplementedError()


class BaseRouter(AbstractRouter, ABC):
    def __init__(
        self, event: "common.EventBuilder", frozen: bool = False, *filters: Filter
    ):
        """
        BaseRouter
        :param frozen: While registering router it can be frozen once.
        """
        self.event = event
        self.frozen = frozen
        self.handlers: PseudoFrozenList[tuple] = PseudoFrozenList[tuple]
        self.filters = filters

    # noinspection PyTypeChecker
    def message_handler(self, *filters: Union[Callable, Filter]):
        return self.register(*(*filters, *self.filters), event=NewMessage)

    # noinspection PyTypeChecker
    def callback_query_handler(self, *filters: Union[Callable, Filter]):
        return self.register(*(*filters, *self.filters), event=CallbackQuery)

    # noinspection PyTypeChecker
    def chat_action_handler(self, *filters: Union[Callable, Filter]):
        return self.register(*(*filters, *self.filters), event=ChatAction)

    # noinspection PyTypeChecker
    def message_edited_handler(self, *filters: Union[Callable, Filter]):
        return self.register(*(*filters, *self.filters), event=MessageEdited)

    def add_router(self, router: "BaseRouter") -> NoReturn:
        if type(self) != type(router):
            raise ValueError(f"Can bind {router!r} into {self!r}")
        for handler in router.handlers:
            self.handlers.append(handler)
        return self, router


class TelethonRouter(BaseRouter):
    """
    Telethon's TgClient.__on__(event) signature followed

    A way to achieve compatibility with telethon event registering.

    Telethon lovers router <3
    """

    def __init__(self, frozen: bool = False):
        super().__init__(None, frozen)

    def register(self, event: "common.EventBuilder"):
        def decorator(f):
            self.handlers.append((f, event, ()))
            return f

        return decorator

    __call__ = on = register

    def non_blocking(self, event: "common.EventBuilder"):
        def decorator(f):
            c = Callback(f)
            c.continue_prop = True
            self.handlers.append((c, event, ()))
            return f

        return decorator


class Router(BaseRouter):
    """
    Garnet's TelegramClient.on(*filters, event=NewMessage) signature followed
    """

    def __init__(
        self, event: "common.EventBuilder" = None, frozen: bool = False, *filters
    ):
        super().__init__(event, frozen, *filters)

    def register(self, *filters: Filter, event: "common.EventBuilder" = None):
        def decorator(f):
            self.handlers.append((f, event or self.event, (*filters, *self.filters)))
            return f

        return decorator

    def __call__(self, *filters: Filter, event: "common.EventBuilder" = None):
        return self.register(*filters, event=event)

    on = __call__

    def non_blocking(self, *filters: Filter, event: "common.EventBuilder" = None):
        def decorator(f):
            c = Callback(f)
            c.continue_prop = True
            self.handlers.append((c, event or self.event, (*filters, *self.filters)))
            return f

        return decorator



# Source: https://github.com/aiogram/aiogram

"""
This module has redis storage for user states based on`aioredis <https://github.com/aio-libs/aioredis>`_ driver
"""

import asyncio
import logging
import typing

import aioredis

from ..jsonlib import json
from .base import BaseStorage

STATE_KEY = "state"
STATE_DATA_KEY = "data"
STATE_BUCKET_KEY = "bucket"


class RedisStorage(BaseStorage):
    """
    Simple Redis-base storage for FSM.
    """

    def __init__(
        self,
        host="localhost",
        port=6379,
        db=None,
        password=None,
        ssl=None,
        loop=None,
        **kwargs,
    ):
        self._host = host
        self._port = port
        self._db = db
        self._password = password
        self._ssl = ssl
        self._loop = loop or asyncio.get_event_loop()
        self._kwargs = kwargs

        self._redis: aioredis.RedisConnection = None
        self._connection_lock = asyncio.Lock(loop=self._loop)

    async def close(self):
        if self._redis and not self._redis.closed:
            self._redis.close()
            del self._redis
            self._redis = None

    async def wait_closed(self):
        if self._redis:
            return await self._redis.wait_closed()
        return True

    async def redis(self) -> aioredis.RedisConnection:
        """
        Get Redis connection
        """
        # Use thread-safe asyncio Lock because this method without that is not safe
        async with self._connection_lock:
            if self._redis is None:
                self._redis = await aioredis.create_connection(
                    (self._host, self._port),
                    db=self._db,
                    password=self._password,
                    ssl=self._ssl,
                    loop=self._loop,
                    **self._kwargs,
                )
        return self._redis

    async def get_record(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
    ) -> typing.Dict:
        """
        Get record from storage

        :param chat:
        :param user:
        :return:
        """
        chat, user = self.check_address(chat=chat, user=user)
        addr = f"fsm:{chat}:{user}"

        conn = await self.redis()
        data = await conn.execute("GET", addr)
        if data is None:
            return {"state": None, "data": {}}
        return json.loads(data)

    async def set_record(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        state=None,
        data=None,
        bucket=None,
    ):
        """
        Write record to storage

        :param bucket:
        :param chat:
        :param user:
        :param state:
        :param data:
        :return:
        """
        if data is None:
            data = {}
        if bucket is None:
            bucket = {}

        chat, user = self.check_address(chat=chat, user=user)
        addr = f"fsm:{chat}:{user}"

        record = {"state": state, "data": data, "bucket": bucket}

        conn = await self.redis()
        await conn.execute("SET", addr, json.dumps(record))

    async def get_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Optional[str]:
        record = await self.get_record(chat=chat, user=user)
        return record["state"]

    async def get_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Dict:
        record = await self.get_record(chat=chat, user=user)
        return record["data"]

    async def set_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        state: typing.Optional[typing.AnyStr] = None,
    ):
        record = await self.get_record(chat=chat, user=user)
        await self.set_record(chat=chat, user=user, state=state, data=record["data"])

    async def set_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
    ):
        record = await self.get_record(chat=chat, user=user)
        await self.set_record(chat=chat, user=user, state=record["state"], data=data)

    async def update_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
        **kwargs,
    ):
        if data is None:
            data = {}
        record = await self.get_record(chat=chat, user=user)
        record_data = record.get("data", {})
        record_data.update(data, **kwargs)
        await self.set_record(
            chat=chat, user=user, state=record["state"], data=record_data
        )

    async def get_states_list(self) -> typing.List[typing.Tuple[int]]:
        """
        Get list of all stored chat's and user's

        :return: list of tuples where first element is chat id and second is user id
        """
        conn = await self.redis()
        result = []

        keys = await conn.execute("KEYS", "fsm:*")
        for item in keys:
            *_, chat, user = item.decode("utf-8").split(":")
            result.append((chat, user))

        return result

    async def reset_all(self, full=True):
        """
        Reset states in DB

        :param full: clean DB or clean only states
        :return:
        """
        conn = await self.redis()

        if full:
            await conn.execute("FLUSHDB")
        else:
            keys = await conn.execute("KEYS", "fsm:*")
            await conn.execute("DEL", *keys)

    def has_bucket(self):
        return True

    async def get_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Dict:
        record = await self.get_record(chat=chat, user=user)
        return record.get("bucket", {})

    async def set_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None,
    ):
        record = await self.get_record(chat=chat, user=user)
        await self.set_record(
            chat=chat,
            user=user,
            state=record["state"],
            data=record["data"],
            bucket=bucket,
        )

    async def update_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None,
        **kwargs,
    ):
        record = await self.get_record(chat=chat, user=user)
        record_bucket = record.get("bucket", {})
        if bucket is None:
            bucket = {}
        record_bucket.update(bucket, **kwargs)
        await self.set_record(
            chat=chat,
            user=user,
            state=record["state"],
            data=record_bucket,
            bucket=bucket,
        )


class RedisStorage2(BaseStorage):
    """
    Busted Redis-base storage for FSM.
    Works with Redis connection pool and customizable keys prefix.

    """

    def __init__(
        self,
        host: str = "localhost",
        port=6379,
        db=None,
        password=None,
        ssl=None,
        pool_size=10,
        loop=None,
        prefix="fsm",
        state_ttl: int = 0,
        data_ttl: int = 0,
        bucket_ttl: int = 0,
        **kwargs,
    ):
        self._host = host
        self._port = port
        self._db = db
        self._password = password
        self._ssl = ssl
        self._pool_size = pool_size
        self._loop = loop or asyncio.get_event_loop()
        self._kwargs = kwargs
        self._prefix = (prefix,)

        self._state_ttl = state_ttl
        self._data_ttl = data_ttl
        self._bucket_ttl = bucket_ttl

        self._redis: aioredis.RedisConnection = None
        self._connection_lock = asyncio.Lock(loop=self._loop)

    async def redis(self) -> aioredis.Redis:
        """
        Get Redis connection
        """
        # Use thread-safe asyncio Lock because this method without that is not safe
        async with self._connection_lock:
            if self._redis is None:
                self._redis = await aioredis.create_redis_pool(
                    (self._host, self._port),
                    db=self._db,
                    password=self._password,
                    ssl=self._ssl,
                    minsize=1,
                    maxsize=self._pool_size,
                    loop=self._loop,
                    **self._kwargs,
                )
        return self._redis

    def generate_key(self, *parts):
        return ":".join(self._prefix + tuple(map(str, parts)))

    async def close(self):
        async with self._connection_lock:
            if self._redis and not self._redis.closed:
                self._redis.close()
                del self._redis
                self._redis = None

    async def wait_closed(self):
        async with self._connection_lock:
            if self._redis:
                return await self._redis.wait_closed()
            return True

    async def get_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None,
    ) -> typing.Optional[str]:
        chat, user = self.check_address(chat=chat, user=user)
        key = self.generate_key(chat, user, STATE_KEY)
        redis = await self.redis()
        return await redis.get(key, encoding="utf8") or None

    async def get_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[dict] = None,
    ) -> typing.Dict:
        chat, user = self.check_address(chat=chat, user=user)
        key = self.generate_key(chat, user, STATE_DATA_KEY)
        redis = await self.redis()
        raw_result = await redis.get(key, encoding="utf8")
        if raw_result:
            return json.loads(raw_result)
        return default or {}

    async def set_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        state: typing.Optional[typing.AnyStr] = None,
    ):
        chat, user = self.check_address(chat=chat, user=user)
        key = self.generate_key(chat, user, STATE_KEY)
        redis = await self.redis()
        if state is None:
            await redis.delete(key)
        else:
            await redis.set(key, state, expire=self._state_ttl)

    async def set_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
    ):
        chat, user = self.check_address(chat=chat, user=user)
        key = self.generate_key(chat, user, STATE_DATA_KEY)
        redis = await self.redis()
        await redis.set(key, json.dumps(data), expire=self._data_ttl)

    async def update_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
        **kwargs,
    ):
        if data is None:
            data = {}
        temp_data = await self.get_data(chat=chat, user=user, default={})
        temp_data.update(data, **kwargs)
        await self.set_data(chat=chat, user=user, data=temp_data)

    def has_bucket(self):
        return True

    async def get_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[dict] = None,
    ) -> typing.Dict:
        chat, user = self.check_address(chat=chat, user=user)
        key = self.generate_key(chat, user, STATE_BUCKET_KEY)
        redis = await self.redis()
        raw_result = await redis.get(key, encoding="utf8")
        if raw_result:
            return json.loads(raw_result)
        return default or {}

    async def set_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None,
    ):
        chat, user = self.check_address(chat=chat, user=user)
        key = self.generate_key(chat, user, STATE_BUCKET_KEY)
        redis = await self.redis()
        await redis.set(key, json.dumps(bucket), expire=self._bucket_ttl)

    async def update_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None,
        **kwargs,
    ):
        if bucket is None:
            bucket = {}
        temp_bucket = await self.get_bucket(chat=chat, user=user)
        temp_bucket.update(bucket, **kwargs)
        await self.set_bucket(chat=chat, user=user, bucket=temp_bucket)

    async def reset_all(self, full=True):
        """
        Reset states in DB

        :param full: clean DB or clean only states
        :return:
        """
        conn = await self.redis()

        if full:
            await conn.flushdb()
        else:
            keys = await conn.keys(self.generate_key("*"))
            await conn.delete(*keys)

    async def get_states_list(self) -> typing.List[typing.Tuple[int]]:
        """
        Get list of all stored chat's and user's

        :return: list of tuples where first element is chat id and second is user id
        """
        conn = await self.redis()
        result = []

        keys = await conn.keys(self.generate_key("*", "*", STATE_KEY), encoding="utf8")
        for item in keys:
            *_, chat, user, _ = item.split(":")
            result.append((chat, user))

        return result

    async def import_redis1(self, redis1):
        await migrate_redis1_to_redis2(redis1, self)


async def migrate_redis1_to_redis2(storage1: RedisStorage, storage2: RedisStorage2):
    """
    Helper for migrating from RedisStorage to RedisStorage2

    :param storage1: instance of RedisStorage
    :param storage2: instance of RedisStorage2
    :return:
    """
    if not isinstance(storage1, RedisStorage):  # better than assertion
        raise TypeError(f"{type(storage1)} is not RedisStorage instance.")
    if not isinstance(storage2, RedisStorage):
        raise TypeError(f"{type(storage2)} is not RedisStorage instance.")

    log = logging.getLogger("aiogram.RedisStorage")

    for chat, user in await storage1.get_states_list():
        state = await storage1.get_state(chat=chat, user=user)
        await storage2.set_state(chat=chat, user=user, state=state)

        data = await storage1.get_data(chat=chat, user=user)
        await storage2.set_data(chat=chat, user=user, data=data)

        bucket = await storage1.get_bucket(chat=chat, user=user)
        await storage2.set_bucket(chat=chat, user=user, bucket=bucket)

        log.info(f"Migrated user {user} in chat {chat}")



from __future__ import annotations
from typing import (
    Union,
    Callable,
    Sequence,
    List,
    Optional,
    NoReturn,
    Any,
    TYPE_CHECKING,
)
import asyncio
import os
import getpass

from telethon.client.telegramclient import TelegramClient as _TelethonTelegramClient

from garnet.events import StopPropagation, NewMessage, Raw

from .events import EventBuilderDict
from .filters import state
from .filters.base import Filter
from .storages import base, file, memory
from .callbacks.base import (
    Callback,
    CURRENT_CLIENT_KEY,
    CALLBACK_STATE_KEY,
)
from .helpers.frozen_list import PseudoFrozenList
from .router import router
from .helpers import ctx

if TYPE_CHECKING:
    from telethon.events import common


class TelegramClient(_TelethonTelegramClient, ctx.ContextInstanceMixin):
    storage: base.BaseStorage = None

    class Env:
        default_session_dsn_key = "SESSION"
        default_api_id_key = "API_ID"
        default_api_hash_key = "API_HASH"
        default_bot_token_key = "BOT_TOKEN"

    def __init__(
        self,
        session: Any,
        api_id: int,
        api_hash: str,
        *args,
        storage: base.BaseStorage = None,
        **kwargs,
    ):
        if storage is not None and isinstance(storage, type):
            raise ValueError("Got wrong value for storage, initialize your storage")

        super().__init__(session, api_id, api_hash, *args, **kwargs)

        self.storage = storage
        self.__bot_token = None

        self.on_start, self.on_finish, self.on_background = (
            PseudoFrozenList[Callable],
            PseudoFrozenList[Callable],
            PseudoFrozenList[Callable],
        )
        self.set_current(self)

    @classmethod
    def from_env(
        cls,
        default_session_dsn: Optional[str] = None,
        default_api_id: Optional[int] = None,
        default_api_hash: Optional[str] = None,
        bot_token: Optional[str] = None,
        storage: base.BaseStorage = None,
        **kwargs,
    ) -> TelegramClient:
        if any(
            kwarg in kwargs for kwarg in ("session", "api_id", "api_hash", "storage")
        ):
            raise ValueError(
                "Don't pass api_id, api_hash, api_hash to .from_env or pass in default_*"
            )
        obj = cls(
            session=os.getenv(cls.Env.default_session_dsn_key, default_session_dsn),
            api_id=os.getenv(cls.Env.default_api_id_key, default_api_id),
            api_hash=os.getenv(cls.Env.default_api_hash_key, default_api_hash),
            storage=storage,
            **kwargs,
        )

        obj.__bot_token = os.getenv(cls.Env.default_bot_token_key, bot_token)
        obj.set_current(obj)

        if isinstance(storage, (file.JSONStorage,)):

            async def close_storage(*_):
                await storage.close()

            obj.on_finish.append(close_storage)

        return obj

    async def start(
        self: "TelegramClient",
        phone: Callable[[], str] = lambda: input(
            "Please enter your phone (or bot token): "
        ),
        password: Callable[[], str] = lambda: getpass.getpass(
            "Please enter your password: "
        ),
        *,
        bot_token: str = None,
        force_sms: bool = False,
        code_callback: Callable[[], Union[str, int]] = None,
        first_name: str = "New User",
        last_name: str = "",
        max_attempts: int = 3,
    ) -> TelegramClient:
        return await _TelethonTelegramClient.start(**locals())

    async def start_as_bot(self, token: str = None) -> TelegramClient:
        self.set_current(self)
        return await self.start(
            bot_token=(token if isinstance(token, str) else self.__bot_token)
        )

    @classmethod
    def make_fsm_key(cls, update, *, _check=True) -> dict:
        try:
            return {"chat": update.chat_id, "user": update.from_id}
        except AttributeError:
            raise Warning(
                f"Standard make_fsm_key method was not designed for {update!r}"
            )

    def current_state(self, *, user, chat):
        return base.FSMContext(storage=self.storage, user=user, chat=chat)

    # end-fsm-key-region

    def bind_routers(self, *routers: router.AbstractRouter) -> NoReturn:
        for router_ in routers:
            if router_.frozen:
                router_.handlers.freeze()

            for handler, event, filters in router_.handlers:
                self.add_event_handler(handler, event or router_.event, *filters)

    def on(
        self, *filters: Union[Callable, Filter], event: common.EventBuilder = NewMessage
    ):
        """
        Decorator used to `add_event_handler` more conveniently.
        """

        def decorator(f):
            self.add_event_handler(f, event, *filters)
            return f

        return decorator

    def add_event_handler(
        self,
        callback: Union[Callable, Callback],
        event=None,
        *filters: Union[Callable, Filter],
    ):
        if filters is not None and not isinstance(filters, Sequence):
            filters = (filters,)

        _has_state_checker = False
        for filter_ in filters:
            if not isinstance(filter_, (Filter, Callable)):
                raise ValueError(
                    f"Got {type(filter_).__qualname__}, expected Filter or Callable"
                )

            if isinstance(filter_, Filter) and filter_.state_op:
                _has_state_checker = True

        if _has_state_checker is False and self.storage is not None:
            filters = list(filters)
            filters.append(~(state == all))
            filters = tuple(filters)

        elif _has_state_checker and self.storage is None:
            self._log[__name__].warning(
                "No storage was set, but state checker was found, using MemoryStorage for now"
            )
            self.storage = memory.MemoryStorage()

        if not isinstance(callback, Callback):
            callback = Callback(callback, *filters)

        if isinstance(event, type):
            event = event()
        elif not event:
            event = Raw()

        self._event_builders.append((event, callback))

    async def _dispatch_update(self, update, others, channel_id, pts_date):
        if not self._entity_cache.ensure_cached(update):
            # We could add a lock to not fetch the same pts twice if we are
            # already fetching it. However this does not happen in practice,
            # which makes sense, because different updates have different pts.
            if self._state_cache.update(update, check_only=True):
                # If the update doesn't have pts, fetching won't do anything.
                # For example, UpdateUserStatus or UpdateChatUserTyping.
                try:
                    await self._get_difference(update, channel_id, pts_date)
                except OSError:
                    pass
                    # We were disconnected, that's okay

        if not self._self_input_peer:
            # Some updates require our own ID, so we must make sure
            # that the event builder has offline access to it. Calling
            # `get_me()` will cache it under `self._self_input_peer`.
            await self.get_me(input_peer=True)

        built = EventBuilderDict(self, update, others)
        filters_cache = {}
        for builder, callback in self._event_builders:
            event = built[type(builder)]
            if not event:
                continue

            if not builder.resolved:
                await builder.resolve(self)

            if not builder.filter(event):
                continue

            filters: List[Filter] = callback.filters
            succeed = True
            context = None

            if self.storage is not None:
                key_maker_from_filters = (
                    list(
                        filter(
                            lambda f_: hasattr(f_, "key_maker")
                            & (f_.key_maker is not None),
                            filters,
                        )
                    )
                    if filters
                    else None
                )

                key_maker = self.make_fsm_key
                if key_maker_from_filters and isinstance(
                    key_maker_from_filters[0].key_maker, Callable
                ):
                    key_maker = key_maker_from_filters[0].key_maker
                context = base.FSMContext(self.storage, **key_maker(event))

            if filters:
                for filter_ in filters:
                    if filter_ in filters_cache:
                        succeed = filters_cache[filter_]

                    if context is not None and filter_.requires_context:
                        if (await filter_.function(event, context)) is False:
                            succeed = False
                            break

                    else:
                        if (
                            (await filter_.function(event))
                            if filter_.awaitable
                            else filter_.function(event)
                        ) is False:
                            succeed = False
                            break

                    filters_cache[filter_] = succeed

            if succeed is False:
                continue

            try:
                kwargs = {}
                if not callback.empty_arg_handler:
                    if callback.requires_client:
                        kwargs[CURRENT_CLIENT_KEY] = self
                    if callback.requires_context:
                        kwargs[CALLBACK_STATE_KEY] = context

                    coro = callback.__call__(event, **kwargs)
                else:
                    coro = callback.__call__(**kwargs)

                if not callback.continue_prop:
                    return await coro
                try:
                    await coro
                except (Exception, ):
                    continue

            except StopPropagation:
                name = getattr(callback, "__name__", repr(callback))
                self._log[__name__].debug(
                    'Event handler "%s" stopped chain of propagation ' "for event %s.",
                    name,
                    type(event).__name__,
                )
                break
            except Exception as e:
                if not isinstance(e, asyncio.CancelledError) or self.is_connected():
                    name = getattr(callback, "__name__", repr(callback))
                    self._log[__name__].exception("Unhandled exception on %s", name)

    def run_tasks(self, *, finishing: bool):
        self.on_start.freeze()
        self.on_background.freeze()
        self.on_finish.freeze()

        async def bulk_await(wait, background):
            # // ordered awaits
            for callback in wait:
                await callback(self)

            for callback in background:
                self.loop.create_task(callback(self))

        if finishing:
            self.loop.run_until_complete(bulk_await(self.on_finish, ()))
        else:
            self.loop.run_until_complete(bulk_await(self.on_start, self.on_background))

    def run_until_disconnected(self):
        self.run_tasks(finishing=False)

        if self.loop.is_running():
            return self._run_until_disconnected()
        try:
            return self.loop.run_until_complete(self._run_until_disconnected())
        except (KeyboardInterrupt, SystemError, SystemExit):
            pass
        finally:
            self.run_tasks(finishing=True)
            self.disconnect()
            self.loop.stop()


__all__ = ("TelegramClient",)



# Source: https://github.com/aiogram/aiogram

import copy
import typing

from .base import BaseStorage


class MemoryStorage(BaseStorage):
    """
    In-memory based states storage.

    This type of storage is not recommended for usage in bots, because you will lost all states after restarting.
    """

    async def wait_closed(self):
        pass

    async def close(self):
        self.data.clear()

    def __init__(self, data: typing.Dict = None):
        self.data = data or {}

    def resolve_address(self, chat, user):
        chat_id, user_id = map(str, self.check_address(chat=chat, user=user))

        if chat_id not in self.data:
            self.data[chat_id] = {}
        if user_id not in self.data[chat_id]:
            self.data[chat_id][user_id] = {"state": None, "data": {}, "bucket": {}}

        return chat_id, user_id

    async def get_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None
    ) -> typing.Optional[str]:
        chat, user = self.resolve_address(chat=chat, user=user)
        return self.data[chat][user]["state"]

    async def get_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[str] = None
    ) -> typing.Dict:
        chat, user = self.resolve_address(chat=chat, user=user)
        return copy.deepcopy(self.data[chat][user]["data"])

    async def update_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None,
        **kwargs
    ):
        if data is None:
            data = {}
        chat, user = self.resolve_address(chat=chat, user=user)
        self.data[chat][user]["data"].update(data, **kwargs)

    async def set_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        state: typing.AnyStr = None
    ):
        chat, user = self.resolve_address(chat=chat, user=user)
        self.data[chat][user]["state"] = state

    async def set_data(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        data: typing.Dict = None
    ):
        chat, user = self.resolve_address(chat=chat, user=user)
        self.data[chat][user]["data"] = copy.deepcopy(data)

    async def reset_state(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        with_data: typing.Optional[bool] = True
    ):
        await self.set_state(chat=chat, user=user, state=None)
        if with_data:
            await self.set_data(chat=chat, user=user, data={})

    def has_bucket(self):
        return True

    async def get_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        default: typing.Optional[dict] = None
    ) -> typing.Dict:
        chat, user = self.resolve_address(chat=chat, user=user)
        return copy.deepcopy(self.data[chat][user]["bucket"])

    async def set_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None
    ):
        chat, user = self.resolve_address(chat=chat, user=user)
        self.data[chat][user]["bucket"] = copy.deepcopy(bucket)

    async def update_bucket(
        self,
        *,
        chat: typing.Union[str, int, None] = None,
        user: typing.Union[str, int, None] = None,
        bucket: typing.Dict = None,
        **kwargs
    ):
        if bucket is None:
            bucket = {}
        chat, user = self.resolve_address(chat=chat, user=user)
        self.data[chat][user]["bucket"].update(bucket, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['garnet',
 'garnet.callbacks',
 'garnet.events',
 'garnet.filters',
 'garnet.functions',
 'garnet.helpers',
 'garnet.router',
 'garnet.storages']

package_data = \
{'': ['*']}

install_requires = \
['Telethon>=1.10.8,<2.0']

setup_kwargs = {
    'name': 'telegram-garnet',
    'version': '0.2.2',
    'description': 'Garnet - cool add-on for Telethon',
    'long_description': '🍷 Garnet\n===================================\n\nGarnet — bot-friendly telethon\n-----------------------------------\n\n.. invisible-content-till-nel\n.. _aioredis: https://github.com/aio-libs/aioredis\n.. _cryptg: https://pypi.org/project/cryptg/\n.. _telethon: https://pypi.org/project/Telethon/\n.. _orjson: https://pypi.org/project/orjson/\n.. _ujson: https://pypi.org/project/ujson/\n.. _hachoir: https://pypi.org/project/hachoir/\n.. _aiohttp: https://pypi.org/project/aiohttp/\n.. _Alex: https://github.com/JrooTJunior\n\n.. image:: https://raw.githubusercontent.com/uwinx/garnet/master/static/pomegranate.jpg\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :target: https://github.com/python/black\n    :alt: aioqiwi-code-style\n\n\nInstall::\n\n    pip install telegram-garnet\n\n\n**Dependencies:**\n    - ``telethon`` - main dependency telethon_\n**Extras:**\n    - ``aioredis`` - redis driver if you use RedisStorage* aioredis_\n    - ``orjson`` || ``ujson`` - RedisStorage/JSONStorage de-&& serialization orjson_ ujson_\n    - ``cryptg``, ``hachoir``, ``aiohttp`` - boost telethon itself cryptg_ hachoir_ aiohttp_\n\n---------------------------------\n🌚 🌝 FSM-Storage types\n---------------------------------\n\n- File - json storage, the main idea behind JSON storage is a custom reload of file and usage of memory session for writing, so the data in json storage not always actual\n\n- Memory - powerful in-memory map<str> based storage, only thing - not persistent\n\n- Redis - (requires aioredis_) - redis is fast key-value storage, if you\'re using your data is durable and persistent\n\n\nPomegranate implements updates dispatching and checks callback\'s filters wrapping all callbacks into ``garnet::Callback`` object\n\n----------------\n😋 Filters\n----------------\n\n``Filter`` object is the essential part of Pomegranate, the do state checking and other stuff.\n\nBasically, it\'s ``func`` from ``MyEventBuilder(func=lambda self: <bool>)`` but a way more complicated and not stored in EventBuilder, it\'s stored in callback object\n\n\nUseful filters\n\n1) 📨 **Messages**\n\n\n`Following examples will include pattern`\n\n\n.. code:: python\n\n    from garnet import MessageText, TelegramClient\n    bot = TelegramClient.from_env().start_as_bot()\n\n    # // our code here //\n\n    bot.run_until_disconnected()\n\n.. code:: python\n\n    # handling /start and /help\n    @bot.on(MessageText.commands("help", "start"))\n    async def cmd_handler(message: custom.Message):\n        await message.reply("Hey there!")\n\n    # handling exact words\n    my_secret = "abcxyz"\n    @bot.on(MessageText.exact(my_secret))\n    async def secret_handler(message: custom.Message):\n        await message.reply("Secret entered correctly! Welcome!")\n\nMessageText or text(``from garnet import text``) includes following comparisons all returning <bool>\n - ``.exact(x)`` -> ``event.raw_text == x``\n - ``.commands(*x)`` -> ``event.raw_text.split()[0][1:] in x``\n - ``.match(x)`` -> ``re.compile(x).match(event.raw_text)``\n - ``.between(*x)`` -> ``event.raw_text in x``\n - ``.isdigit()`` -> ``(event.raw_text or "").isdigit()``\n - ``.startswith(x)`` -> ``event.raw_text.startswith(x)``\n\n\n\n2) 👀 **CurrentState class**  [``from garnet import CurrentState``]\n\nOnce great minds decided that state checking will be in filters without adding ``state`` as handler decorator parameter and further storing state in ``callback.(arg)``\n``CurrentState`` class methods return ``Filter``. There are two problems that Filter object really solves, ``Filter``\'s function can be any kind of callable(async,sync), filters also have a flag ``requires_context``, FSMProxy is passed if true\n\nSee `FSM example <https://github.com/uwinx/garnet/blob/master/examples/fsm.py>`_ to understand how CurrentState works\n\nIncludes following methods all returning <bool>\n - ``.exact(x)`` or ``CurrentState == x`` -> ``await context.get_state() == x``\n - ``CurrentState == [x, y, z]`` -> ``await context.get_state() in [x, y, z]``\n - ``CurrentState == all`` or ``CurrentState == any`` -> ``await context.get_state() is not None``\n\n\n3) 🦔 Custom **Filter**\n\nIf you want to write your own filter, do it.\n\n\n.. code:: python\n\n    from garnet import Filter, FSMProxy\n\n    async def myFunc(event): ...\n    async def myFuncContextRequires(event, context: FSMProxy): ...\n    def normal_func(event): ...\n\n    @bot.on(Filter(normal_func), Filter(myFunc), Filter(myFuncContextRequires, requires_context=True))\n    async def handler(event, context: FSMProxy): ...\n    # same as\n    @bot.on(normal_func, myFunc, Filter(myFuncContextRequires, requires_context=True))\n    async def handler(event): ...\n\nSo the handler can take strict ``context`` argument and also ignore it\n\n-----------------------\nOn start|finish\n-----------------------\n\n``garnet::TelegramClient`` contains three lists on_start on_background and on_finish, their instance is ``PseudoFrozenList`` which freezes at calling ``.run_until_disconnected``\n``PseudoFrozenList`` has three main methods::\n\n    .append(*items)\n    .remove(*items)\n    .freeze()\n    .__call__(func)   # for shiny decorator\n\n``items`` in case of TelegramClient means unpacked container of async-defined functions taking on position arguments\n\nUsage example:\n\n.. code-block:: python\n\n    # my_module.py\n    class MyPostgresDatabase:\n        ...\n        async def close_pool(self, bot): await self.pool.safe_close()\n        async def open_pool(self, bot): await self.pool.open_conn_pool()\n\n    # garnethon.py\n    from garnet import TelegramClient\n    from my_module import MyPostgresDatabase\n\n    db = MyPostgresDatabase()\n    bot = TelegramClient.from_env().start_as_bot()\n    bot.on_start.append(db.open_pool)\n    bot.on_finish.append(db.close_pool)\n    ...\n\n    @bot.on_background\n    async def xyz(cl: TelegramClient):\n        while True:\n           ...\n\n    bot.run_until_connected()\n\n\n-------------------------------------------------\n📦 Router and Migrating to garnet using Router\n-------------------------------------------------\n\nThink of router as just a dummy container of handlers(callbacks)\n\n`garnet::router::Router` may be helpful if you have telethon\'s `event.register` registered handlers. One thing: Router, I believe, is correct and more obvious way of registering event handlers. Example:\n\n**Migrate from telethon to garnet, also use for bot.on cases(soon better example)**\n\n.. code-block:: python\n\n    # my_handlers.py\n\n    # telethon register(bad) will raise Warning in garnet\n    from telethon import events\n\n    @events.register(event_type)\n    async def handler(event): ...\n\n    # garnet\'s telethon-like router\n    from garnet.router import TelethonRouter\n\n    router = TelethonRouter()\n\n    @router(event_type)\n    async def handler(event): ...\n\n\n\nThe advantage of routers is evidence of registering handlers when you have module-separated handlers. `events.register` was doing well, but blindly importing modules to register handlers and don\'t use them(modules) doesn\'t seem like a good idea.\n\n\nExample of registering router in bot application\n\n\n.. code-block:: python\n\n    # handlers/messages.py\n    from garnet.router import Router\n\n    router = Router()\n\n    @router()\n    async def handler(event): ...\n\n    # handlers/cb_query.py\n    from garnet.events import CallbackQuery\n    from garnet.router import Router\n\n    router = Router()\n\n    @router(event=CallbackQuery())\n    async def handler(event): ...\n\n    # entry.py ()\n    from garnet import TelegramClient\n\n    from handlers import messages, cb_query\n\n    tg = TelegramClient.from_env().start_as_bot()\n    tg.bind_routers(messages, cb_query)\n    ...\n\n`TelethonRouter` and `Router` both have following remarkable methods:\n\n::\n\n    .message_handler(*filters)\n    .callback_query_handler(*filters)\n    .chat_action_handler(*filters)\n    .message_edited_handler(*filters)\n    .album_handler(*filters)\n\n--------------------\n🍬 Context magic\n--------------------\n\nOne of the sweetest parts of garnet. Using `contextvars` we reach incredibly beautiful code :D\n*this is not FSMContext don\'t confuse with context magic provided by contextvars*\n\nAs an example, bot that doesn\'t requires `TelegramClient` to answer messages directly.\n\n.. code-block:: python\n\n    from garnet.functions.messages import reply, message, respond\n\n    @bot.message_handler()\n    async def handler():\n        # message() - function to get current Message event\n        await message().respond("ok")\n        await message().reply("ok")\n        # the same result, but shortcuts\n        await respond("ok")\n        await reply("Ok")\n\n\n\n-----------------\nWhat\'s more ❓\n-----------------\n\nClass-based handlers are also can be implemented with garnet conveniently. Use your imagination and ``garnet::callbacks::base::Callback`` as a parent class\n\nAwesome bitwise operation supported filters(I highly recommend to use them)::\n\n    # & (conjunction), | (disjunction), ~ (inversion), ^ (exclusive disjunction)\n    # also: ==, != (idk why)\n    @bot.on(MessageText.exact(".") | MessageText.exact(".."))\n\n\n``Len`` attribute in ``MessageText`` which has cmp methods::\n\n\n    @bot.on((MessageText.Len <= 14) | (MessageText.Len >= 88))\n\n\nUsing `client = TelegramClient.start` assignment and start client on the fly, make annotation or typing.cast to have better hints.\n\n---------------\nAbout\n---------------\n\nYou can find me in tg by `@martin_winks <https://telegram.me/martin_winks>`_ and yeah I receive donates as well as all contributors do(support `lonamiwebs <http://paypal.me/lonamiwebs>`_ and `JRootJunior <https://opencollective.com/aiogram/organization/0/website>`_).\n\n\n---------------------\n🤗 Credits\n---------------------\n\nFinite-state machine was ported from cool BotAPI library \'aiogram\', special thanks to Alex_\n',
    'author': 'Martin Winks',
    'author_email': 'cat@snejugal.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/uwinx/garnet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



# Source: https://github.com/aiogram/aiogram
# MODIFIED.

import asyncio
import pathlib
import typing
from concurrent.futures import ThreadPoolExecutor

from ..jsonlib import json
from .memory import MemoryStorage

executor = ThreadPoolExecutor()
loop = asyncio.get_event_loop()


def _create_json_file(path: pathlib.Path):
    """
    Since `orJSON` does not provide .dump write by hand.
    """
    with path.open("w", encoding="utf-8") as f:
        f.write("{}")


class JSONStorage(MemoryStorage):
    """
    JSON File storage based on MemoryStorage
    """

    def __init__(self, path: typing.Union[pathlib.Path, str]):
        """
        Creates if not found, saves if process was closed properly.
        Could be more reliable.
        :param path: file path
        """
        self.path = path = pathlib.Path(path)
        try:
            data = self.__read(path)
        except FileNotFoundError:
            _create_json_file(path)
            data = self.__read(path)

        super().__init__(data)

    # noinspection PyMethodMayBeStatic
    def __read(self, path: pathlib.Path):
        with path.open("r", encoding="utf-8") as f:
            return json.loads(f.read())

    def _save(self, path: pathlib.Path):
        with path.open("w", encoding="utf-8") as f:
            return f.write(json.dumps(self.data).decode())

    async def save(self, close: bool = False):
        if self.data:
            await loop.run_in_executor(executor, self.save, self.path)
        if close:
            await super().close()



import typing

from telethon import hints, types, custom

from garnet.events import NewMessage


def message() -> custom.Message:
    # noinspection PyTypeChecker
    return NewMessage.Event.current()  # type: ignore


# noinspection PyPep8Naming
class current(object):  # type: ignore
    @property
    def chat_id(self) -> typing.Optional[int]:
        """
        get current chat's ID
        """
        return message().chat_id

    @property
    def chat(self) -> typing.Union[types.User, types.Chat, types.Channel]:
        """
        get current chat
        """
        return message().chat

    @property
    def fmt_text(self) -> typing.Optional[str]:
        """
        Get formatted text with respect to current parse_mode
        e.g. <code>...</code>
        """
        return message().text

    @property
    def text(self) -> typing.Optional[str]:
        """
        Get RAW text from update, text without formatting
        """
        return message().raw_text


async def respond(
    message_: "hints.MessageLike" = "",
    *,
    reply_to: "typing.Union[int, types.Message]" = None,
    parse_mode: typing.Optional[str] = (),
    link_preview: bool = True,
    file: "typing.Union[hints.FileLike, typing.Sequence[hints.FileLike]]" = None,
    force_document: bool = False,
    clear_draft: bool = False,
    buttons: "hints.MarkupLike" = None,
    silent: bool = None,
    schedule: "hints.DateLike" = None
) -> "types.Message":

    _l = locals()
    _l["message"] = _l.pop("message_")
    return await message().respond(**locals())


async def reply(
    message_: "hints.MessageLike" = "",
    *,
    reply_to: "typing.Union[int, types.Message]" = None,
    parse_mode: typing.Optional[str] = (),
    link_preview: bool = True,
    file: "typing.Union[hints.FileLike, typing.Sequence[hints.FileLike]]" = None,
    force_document: bool = False,
    clear_draft: bool = False,
    buttons: "hints.MarkupLike" = None,
    silent: bool = None,
    schedule: "hints.DateLike" = None
) -> "types.Message":

    _l = locals()
    _l["message"] = _l.pop("message_")
    return await message().reply(**_l)


__all__ = ("current", "reply", "respond", "message")
